# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Pandas accessor."""

import os
import unittest

# pylint: disable=unused-import
import tfrecorder

from tfrecorder import test_utils


class DataFrameAccessor(unittest.TestCase):
  """UnitTests For DataFrame Accessor."""

  def setUp(self):
    self.test_df = test_utils.get_test_df()
    self.output_dir = '/tmp/train'
    os.makedirs(self.output_dir, exist_ok=True)

  def test_accessor(self):
    """Tests pandas accessor."""

    r = self.test_df.tensorflow.to_tfr(
        runner='DirectRunner', output_dir=self.output_dir)
    self.assertTrue('metrics' in r)


if __name__ == '__main__':
  unittest.main()



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates a pandas DataFrame accessor for TFRecorder.

accessor.py contains TFRecorderAccessor which provides a pandas DataFrame
accessor.  This accessor allows us to inject the to_tfr() function into
pandas DataFrames.
"""

from typing import Any, Dict, Optional, Union
import pandas as pd
from IPython.core import display

from tfrecorder import client
from tfrecorder import constants


@pd.api.extensions.register_dataframe_accessor('tensorflow')
class TFRecorderAccessor:
  """DataFrame Accessor class for TFRecorder."""

  def __init__(self, pandas_obj):
    self._df = pandas_obj

  # pylint: disable=too-many-arguments
  def to_tfr(
      self,
      output_dir: str,
      runner: str = 'DirectRunner',
      project: Optional[str] = None,
      region: Optional[str] = None,
      tfrecorder_wheel: Optional[str] = None,
      dataflow_options: Union[Dict[str, Any], None] = None,
      job_label: str = 'to-tfr',
      compression: Optional[str] = 'gzip',
      num_shards: int = 0) -> Dict[str, Any]:
    """TFRecorder Pandas Accessor.

    TFRecorder provides an easy interface to create image-based tensorflow
    records from a dataframe containing GCS locations of the images and labels.

    Usage:
      import tfrecorder

      df.tfrecorder.to_tfr(
          output_dir='gcs://foo/bar/train',
          runner='DirectRunner',
          compression='gzip',
          num_shards=10)

    Args:
      output_dir: Local directory or GCS Location to save TFRecords to.
      runner: Beam runner. Can be DirectRunner or  DataFlowRunner.
      project: GCP project name (Required if DataFlowRunner).
      region: GCP region name (Required if DataFlowRunner).
      tfrecorder_wheel: Path to the tfrecorder wheel DataFlow will run.
        (create with 'python setup.py sdist' or
        'pip download tfrecorder --no-deps')
      dataflow_options: Optional dictionary containing DataFlow options.
      job_label: User supplied description for the beam job name.
      compression: Can be 'gzip' or None for no compression.
      num_shards: Number of shards to divide the TFRecords into. Default is
          0 = no sharding.
    Returns:
      job_results: A dictionary of job results.
    """
    display.display(
        display.HTML(
            '<b>Logging output to /tmp/{} </b>'.format(constants.LOGFILE)))

    r = client.create_tfrecords(
        self._df,
        output_dir=output_dir,
        runner=runner,
        project=project,
        region=region,
        tfrecorder_wheel=tfrecorder_wheel,
        dataflow_options=dataflow_options,
        job_label=job_label,
        compression=compression,
        num_shards=num_shards)
    return r



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFRecorder Beam Pipeline.

This file implements the full Beam pipeline for TFRecorder.
"""

from typing import Any, Dict, Generator, Union

import functools
import logging
import os

import apache_beam as beam
import pandas as pd
import tensorflow_transform as tft
from tensorflow_transform import beam as tft_beam

from tfrecorder import beam_image
from tfrecorder import common
from tfrecorder import constants


def _get_job_name(job_label: str = None) -> str:
  """Returns Beam runner job name.

  Args:
    job_label: A user defined string that helps define the job.

  Returns:
    A job name compatible with apache beam runners, including a time stamp to
      insure uniqueness.
  """

  job_name = 'tfrecorder-' + common.get_timestamp()
  if job_label:
    job_label = job_label.replace('_', '-')
    job_name += '-' + job_label

  return job_name


def _get_job_dir(output_path: str, job_name: str) -> str:
  """Returns Beam processing job directory."""

  return os.path.join(output_path, job_name)


def _get_pipeline_options(
    runner: str,
    job_name: str,
    job_dir: str,
    project: str,
    region: str,
    tfrecorder_wheel: str,
    dataflow_options: Union[Dict[str, Any], None]
    ) -> beam.pipeline.PipelineOptions:
  """Returns Beam pipeline options."""

  options_dict = {
      'runner': runner,
      'staging_location': os.path.join(job_dir, 'staging'),
      'temp_location': os.path.join(job_dir, 'tmp'),
      'job_name': job_name,
      'teardown_policy': 'TEARDOWN_ALWAYS',
      'save_main_session': True,
      'pipeline_type_check': False,
  }

  if project:
    options_dict['project'] = project
  if region:
    options_dict['region'] = region
  if runner == 'DataflowRunner':
    options_dict['extra_packages'] = tfrecorder_wheel
  if dataflow_options:
    options_dict.update(dataflow_options)

  return beam.pipeline.PipelineOptions(flags=[], **options_dict)


def _partition_fn(
    element: Dict[str, str],
    unused_num_partitions: int = -1) -> int:
  """Returns index used to partition an element from a PCollection."""
  del unused_num_partitions
  dataset_type = element[constants.SPLIT_KEY].decode('utf-8')
  try:
    index = constants.SPLIT_VALUES.index(dataset_type)
  except ValueError as e:
    logging.warning('Unable to index dataset type %s: %s.',
                    dataset_type, str(e))
    index = constants.DISCARD_INDEX
  return index

def _get_write_to_tfrecord(output_dir: str,
                           prefix: str,
                           compress: bool = True,
                           num_shards: int = 0) \
                           -> beam.io.tfrecordio.WriteToTFRecord:
  """Returns `beam.io.tfrecordio.WriteToTFRecord` object.

  This configures a Beam sink to output TFRecord files.

  Args:
    output_dir: Directory to output TFRecord files.
    prefix: TFRecord file prefix.
    compress: If True, GZip compress TFRecord files.
    num_shards: Number of file shards to split the TFRecord data.
  """

  path = os.path.join(output_dir, prefix)
  suffix = '.tfrecord'
  if compress:
    compression_type = 'gzip'
    suffix += '.gz'
  else:
    compression_type = 'uncompressed'

  return beam.io.tfrecordio.WriteToTFRecord(
      path,
      file_name_suffix=suffix,
      compression_type=compression_type,
      num_shards=num_shards,
  )

def _preprocessing_fn(inputs, integer_label: bool = False):
  """TensorFlow Transform preprocessing function."""

  outputs = inputs.copy()

  if not integer_label:
    # Integerize string labels, if present.
    outputs[constants.LABEL_KEY] = tft.compute_and_apply_vocabulary(
        outputs[constants.LABEL_KEY])

  return outputs


# pylint: disable=abstract-method

class ToCSVRows(beam.DoFn):
  """Adds image to PCollection."""

  def __init__(self):
    """Constructor."""
    super().__init__()
    self.row_count = beam.metrics.Metrics.counter(self.__class__, 'row_count')


  # pylint: disable=unused-argument
  # pylint: disable=arguments-differ
  def process(
      self,
      element: Dict[str, Any]
      ) -> Generator[Dict[str, Any], None, None]:
    """Loads image and creates image features.

    This DoFn extracts an image being stored on local disk or GCS and
    yields a base64 encoded image, the image height, image width, and channels.
    """
    element = ','.join([str(item) for item in element])
    self.row_count.inc()
    yield element



# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def build_pipeline(
    df: pd.DataFrame,
    job_label: str,
    runner: str,
    project: str,
    region: str,
    output_dir: str,
    compression: str,
    num_shards: int,
    tfrecorder_wheel: str,
    dataflow_options: dict,
    integer_label: bool) -> beam.Pipeline:
  """Runs TFRecorder Beam Pipeline.

  Args:
    df: Pandas DataFrame
    job_label: User description for the beam job.
    runner: Beam Runner: (e.g. DataflowRunner, DirectRunner).
    project: GCP project ID (if DataflowRunner)
    region: GCP compute region (if DataflowRunner)
    output_dir: GCS or Local Path for output.
    compression: gzip or None.
    num_shards: Number of shards.
    tfrecorder_wheel: Path to TFRecorder wheel for DataFlow
    dataflow_options: Dataflow Runner Options (optional)
    integer_label: Flags if label is already an integer.

  Returns:
    beam.Pipeline

  Note: These inputs must be validated upstream (by client.create_tfrecord())
  """

  job_name = _get_job_name(job_label)
  job_dir = _get_job_dir(output_dir, job_name)
  options = _get_pipeline_options(
      runner,
      job_name,
      job_dir,
      project,
      region,
      tfrecorder_wheel,
      dataflow_options)

  #with beam.Pipeline(runner, options=options) as p:
  p = beam.Pipeline(options=options)
  with tft_beam.Context(temp_dir=os.path.join(job_dir, 'tft_tmp')):

    converter = tft.coders.CsvCoder(constants.IMAGE_CSV_COLUMNS,
                                    constants.IMAGE_CSV_METADATA.schema)

    extract_images_fn = beam_image.ExtractImagesDoFn(constants.IMAGE_URI_KEY)
    flatten_rows = ToCSVRows()

    # Each element in the image_csv_data PCollection will be a dict
    # including the image_csv_columns and the image features created from
    # extract_images_fn.
    image_csv_data = (
        p
        | 'ReadFromDataFrame' >> beam.Create(df.values.tolist())
        | 'ToCSVRows' >> beam.ParDo(flatten_rows)
        | 'DecodeCSV' >> beam.Map(converter.decode)
        | 'ReadImage' >> beam.ParDo(extract_images_fn)
    )

    # Split dataset into train and validation.
    train_data, val_data, test_data, discard_data = (
        image_csv_data | 'SplitDataset' >> beam.Partition(
            _partition_fn, len(constants.SPLIT_VALUES))
    )

    train_dataset = (train_data, constants.RAW_METADATA)
    val_dataset = (val_data, constants.RAW_METADATA)
    test_dataset = (test_data, constants.RAW_METADATA)

    # TensorFlow Transform applied to all datasets.
    preprocessing_fn = functools.partial(
        _preprocessing_fn,
        integer_label=integer_label)
    transformed_train_dataset, transform_fn = (
        train_dataset
        | 'AnalyzeAndTransformTrain' >> tft_beam.AnalyzeAndTransformDataset(
            preprocessing_fn))

    transformed_train_data, transformed_metadata = transformed_train_dataset
    transformed_data_coder = tft.coders.ExampleProtoCoder(
        transformed_metadata.schema)

    transformed_val_data, _ = (
        (val_dataset, transform_fn)
        | 'TransformVal' >> tft_beam.TransformDataset()
    )

    transformed_test_data, _ = (
        (test_dataset, transform_fn)
        | 'TransformTest' >> tft_beam.TransformDataset()
    )

    # Sinks for TFRecords and metadata.
    tfr_writer = functools.partial(_get_write_to_tfrecord,
                                   output_dir=job_dir,
                                   compress=compression,
                                   num_shards=num_shards)

    _ = (
        transformed_train_data
        | 'EncodeTrainData' >> beam.Map(transformed_data_coder.encode)
        | 'WriteTrainData' >> tfr_writer(prefix='train'))

    _ = (
        transformed_val_data
        | 'EncodeValData' >> beam.Map(transformed_data_coder.encode)
        | 'WriteValData' >> tfr_writer(prefix='val'))

    _ = (
        transformed_test_data
        | 'EncodeTestData' >> beam.Map(transformed_data_coder.encode)
        | 'WriteTestData' >> tfr_writer(prefix='test'))

    _ = (
        discard_data
        | 'DiscardDataWriter' >> beam.io.WriteToText(
            os.path.join(job_dir, 'discarded-data')))

    # Output transform function and metadata
    _ = (transform_fn | 'WriteTransformFn' >> tft_beam.WriteTransformFn(
        job_dir))

    # Output metadata schema
    _ = (transformed_metadata | 'WriteMetadata' >> tft_beam.WriteMetadata(
        job_dir, pipeline=p))

  return p



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Global constants."""

import logging

import frozendict
import tensorflow as tf
from tensorflow_transform.tf_metadata import dataset_metadata
from tensorflow_transform.tf_metadata import schema_utils


SPLIT_KEY = 'split'
SPLIT_VALUES = ['TRAIN', 'VALIDATION', 'TEST', 'DISCARD']
DISCARD_INDEX = SPLIT_VALUES.index('DISCARD')
IMAGE_URI_KEY = 'image_uri'
LABEL_KEY = 'label'
IMAGE_CSV_COLUMNS = [SPLIT_KEY, IMAGE_URI_KEY, LABEL_KEY]

IMAGE_CSV_FEATURE_SPEC = frozendict.FrozenOrderedDict({
    SPLIT_KEY: tf.io.FixedLenFeature([], tf.string),
    IMAGE_URI_KEY: tf.io.FixedLenFeature([], tf.string),
    LABEL_KEY: tf.io.FixedLenFeature([], tf.string),
})

RAW_FEATURE_SPEC = frozendict.FrozenOrderedDict({
    SPLIT_KEY: tf.io.FixedLenFeature([], tf.string),
    LABEL_KEY: tf.io.FixedLenFeature([], tf.string),
    'image_name': tf.io.FixedLenFeature([], tf.string),
    'image': tf.io.FixedLenFeature([], tf.string),
    'image_height': tf.io.FixedLenFeature([], tf.int64),
    'image_width': tf.io.FixedLenFeature([], tf.int64),
    'image_channels':  tf.io.FixedLenFeature([], tf.int64)
})

IMAGE_CSV_METADATA = dataset_metadata.DatasetMetadata(
    schema_utils.schema_from_feature_spec(IMAGE_CSV_FEATURE_SPEC))

RAW_METADATA = dataset_metadata.DatasetMetadata(
    schema_utils.schema_from_feature_spec(RAW_FEATURE_SPEC))

LOGFILE = 'tfrecorder-beam.log'
LOGLEVEL = logging.INFO

CONSOLE_DATAFLOW_URI = "https://console.cloud.google.com/dataflow/jobs/"



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utility functions."""

from datetime import datetime
import os

import tensorflow as tf

from tfrecorder import constants


def get_timestamp() -> str:
  """Returns current date and time as formatted string."""
  return datetime.now().strftime('%Y%m%d-%H%M%S')


def copy_logfile_to_gcs(logfile: str, output_dir: str):
  """Copies a logfile from local to gcs storage."""
  try:
    with open(logfile, 'r') as log_reader:
      out_log = os.path.join(output_dir, constants.LOGFILE)
      with tf.io.gfile.GFile(out_log, 'w') as gcs_logfile:
        log = log_reader.read()
        gcs_logfile.write(log)
  except FileNotFoundError as e:
    raise FileNotFoundError("Unable to copy log file {} to gcs.".format(
        e.filename)) from e



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for beam_pipeline."""

import unittest
from unittest import mock

import apache_beam as beam
import tensorflow as tf

from tfrecorder import beam_pipeline


# pylint: disable=protected-access

class BeamPipelineTests(unittest.TestCase):
  """Tests for beam_pipeline.py"""

  def test_processing_fn_with_int_label(self):
    'Test preprocessing fn with integer label.'
    element = {
        'split': 'TRAIN',
        'image_uri': 'gs://foo/bar.jpg',
        'label': 1}
    result = beam_pipeline._preprocessing_fn(element, integer_label=True)
    self.assertEqual(element, result)

  @mock.patch('tfrecorder.beam_pipeline.tft')
  def test_processing_fn_with_string_label(self, mock_transform):
    'Test preprocessing fn with string label.'
    mock_transform.compute_and_apply_vocabulary.return_value = tf.constant(
        0, dtype=tf.int64)
    element = {
        'split': 'TRAIN',
        'image_uri': 'gs://foo/bar.jpg',
        'label': tf.constant('cat', dtype=tf.string)}
    result = beam_pipeline._preprocessing_fn(element, integer_label=False)
    result['label'] = result['label'].numpy()
    self.assertEqual(0, result['label'])

  def test_write_to_tfrecord(self):
    """Test _write_to_tfrecord() fn."""
    tfr_writer = beam_pipeline._get_write_to_tfrecord(
        output_dir='tmp',
        prefix='foo',
        compress=True,
        num_shards=2)
    self.assertIsInstance(tfr_writer, beam.io.tfrecordio.WriteToTFRecord)

  def test_partition_fn(self):
    """Test the partition function."""

    test_data = {
        'split': 'update_me',
        'image_uri': 'gs://foo/bar0.jpg',
        'label': 1}

    for i, part in enumerate(['TRAIN', 'VALIDATION', 'TEST', 'FOO']):
      test_data['split'] = part.encode('utf-8')
      index = beam_pipeline._partition_fn(test_data)

      self.assertEqual(
          index, i,
          '{} should be index {} but was index {}'.format(part, i, index))


if __name__ == '__main__':
  unittest.main()



# Lint as: python3

# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests `check.py`."""

import functools
import os
import tempfile
import unittest

import mock
import pandas as pd
from pandas import testing as pdt
import tensorflow as tf

from tfrecorder import beam_image
from tfrecorder import check
from tfrecorder import constants
from tfrecorder import test_utils


# pylint: disable=protected-access

class ReadTFRecordsTest(unittest.TestCase):
  """Tests `_read_tfrecords`."""

  def setUp(self):
    self.tfrecords_dir = os.path.join(test_utils.TEST_DIR, 'sample_tfrecords')

  def test_valid_compressed_gzip(self):
    """Tests valid case using GZIP compression."""

    # Use list of file pattern strings to maintain train, validation, test
    # order.
    file_pattern = [
        os.path.join(self.tfrecords_dir, '{}*.tfrecord.gz'.format(f))
        for f in ['train, validation, test']]

    compression_type = 'GZIP'
    actual = check._read_tfrecords(
        file_pattern, self.tfrecords_dir, compression_type)

    expected_csv = os.path.join(test_utils.TEST_DIR, 'data.csv')
    expected = tf.data.experimental.make_csv_dataset(
        expected_csv, batch_size=1, label_name=None, num_epochs=1,
        shuffle=False)

    for a, e in zip(actual, expected):
      self.assertCountEqual(a.keys(), constants.RAW_FEATURE_SPEC)
      for key in constants.IMAGE_CSV_FEATURE_SPEC:
        self.assertEqual(a[key], e[key])

  def test_error_invalid_file_pattern(self):
    """Tests error case where file pattern is invalid."""

    file_pattern = 'gs://path/to/memes/folder'
    with self.assertRaises(tf.errors.OpError):
      check._read_tfrecords(file_pattern)


class CheckTFRecordsTest(unittest.TestCase):
  """Tests `check_tfrecords`."""

  def setUp(self):
    """Test setup."""

    image_height = 40
    image_width = 30
    image_channels = 3
    image_fn = functools.partial(
        test_utils.make_random_image, image_height, image_width,
        image_channels)

    data = test_utils.get_test_data()
    num_records = len(data[constants.IMAGE_URI_KEY])
    image_uris = data.pop(constants.IMAGE_URI_KEY)
    data['image_name'] = [os.path.split(uri)[-1] for uri in image_uris]
    data.update({
        'image': [beam_image.encode(image_fn())
                  for _ in range(num_records)],
        'image_height': [image_height] * num_records,
        'image_width': [image_width] * num_records,
        'image_channels': [image_channels] * num_records,
    })
    self.num_records = num_records
    self.data = data
    self.dataset = tf.data.Dataset.from_tensor_slices(self.data)

  @mock.patch.object(check, '_read_tfrecords', autospec=True)
  def test_valid_records(self, mock_fn):
    """Tests valid case on reading multiple records."""

    file_pattern = 'gs://path/to/tfrecords/*'
    mock_fn.return_value = self.dataset
    num_records = len(self.data['image'])

    with tempfile.TemporaryDirectory(dir='/tmp') as dir_:
      actual_dir = check.check_tfrecords(
          file_pattern, num_records=num_records, output_dir=dir_)
      self.assertTrue('check-tfrecords-' in actual_dir)

      actual_csv = os.path.join(actual_dir, 'data.csv')
      self.assertTrue(os.path.exists(actual_csv))

      _ = self.data.pop('image')

      # Check output CSV
      actual_df = pd.read_csv(actual_csv)
      expected_df = pd.DataFrame(self.data)
      pdt.assert_frame_equal(actual_df, expected_df)

      # Check output images
      actual_image_files = [
          f for f in os.listdir(actual_dir) if f.endswith('.jpg')]
      expected_image_files = self.data['image_name']
      self.assertCountEqual(actual_image_files, expected_image_files)


if __name__ == '__main__':
  unittest.main()
